@@ -1,6 +1,6 @@
 /* 
    neon test suite
-   Copyright (C) 2002-2007, Joe Orton <joe@manyfish.co.uk>
+   Copyright (C) 2002-2009, Joe Orton <joe@manyfish.co.uk>
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -45,10 +45,14 @@
 #error SSL not supported
 #endif
 
+#include "ne_pkcs11.h"
 
 #define SERVER_CERT "server.cert"
+#define CA2_SERVER_CERT "ca2server.pem"
 #define CA_CERT "ca/cert.pem"
 
+#define P12_PASSPHRASE "foobar"
+
 #define SERVER_DNAME "Neon QA Dept, Neon Hackers Ltd, " \
                      "Cambridge, Cambridgeshire, GB"
 #define CACERT_DNAME "Random Dept, Neosign, Oakland, California, US"
@@ -59,6 +63,8 @@ static char *server_key = NULL;
 static ne_ssl_certificate *def_ca_cert = NULL, *def_server_cert;
 static ne_ssl_client_cert *def_cli_cert;
 
+static char *nul_cn_fn;
+
 static int check_dname(const ne_ssl_dname *dn, const char *expected,
                        const char *which);
 
@@ -71,7 +77,7 @@ struct ssl_server_args {
     /* client cert handling: */
     int require_cc; /* require a client cert if non-NULL */
     const char *ca_list; /* file of CA certs to verify client cert against */
-    const char *send_ca; /* file of CA certs to send in client cert request */
+    int fail_silently; /* exit with success if handshake fails */
     
     /* session caching: */
     int cache; /* use the session cache if non-zero */
@@ -82,6 +88,8 @@ struct ssl_server_args {
     int count; /* internal use. */
 
     int use_ssl2; /* force use of SSLv2 only */
+
+    const char *key;
 };
 
 /* default response string if args->response is NULL */
@@ -93,6 +101,7 @@ static int ssl_server(ne_socket *sock, void *userdata)
     struct ssl_server_args *args = userdata;
     int ret;
     char buf[BUFSIZ];
+    const char *key;
     static ne_ssl_context *ctx = NULL;
 
     if (ctx == NULL) {
@@ -102,20 +111,25 @@ static int ssl_server(ne_socket *sock, void *userdata)
 
     ONV(ctx == NULL, ("could not create SSL context"));
 
-    ONV(ne_ssl_context_keypair(ctx, args->cert, server_key),
+    key = args->key ? args->key : server_key;
+    NE_DEBUG(NE_DBG_HTTP, "SSL server init with keypair (%s, %s).\n",
+             args->cert, key);
+             
+    ONV(ne_ssl_context_keypair(ctx, args->cert, key),
         ("failed to load server keypair: ..."));
 
-    NE_DEBUG(NE_DBG_HTTP, "using server cert %s\n", args->cert);
-
     if (args->require_cc && !args->ca_list) {
         args->ca_list = CA_CERT;
     }
 
-    ne_ssl_context_set_verify(ctx, args->require_cc, args->send_ca,
-                              args->ca_list);
+    ne_ssl_context_set_verify(ctx, args->require_cc, 
+                              args->ca_list, args->ca_list);
 
-    ONV(ne_sock_accept_ssl(sock, ctx),
-        ("SSL accept failed: %s", ne_sock_error(sock)));
+    ret = ne_sock_accept_ssl(sock, ctx);
+    if (ret && args->fail_silently) {
+        return 0;
+    }
+    ONV(ret, ("SSL accept failed: %s", ne_sock_error(sock)));
 
     args->count++;
 
@@ -244,13 +258,26 @@ static int init(void)
     
     /* tests for the encrypted client cert, client.p12 */
     def_cli_cert = ne_ssl_clicert_read("client.p12");
-    ONN("could not load client.p12", def_cli_cert == NULL);
+    if (def_cli_cert == NULL) {
+        t_context("could not load client.p12");
+        return FAILHARD;
+    }
 
-    ONN("client.p12 is not encrypted!?", 
-        !ne_ssl_clicert_encrypted(def_cli_cert));
-    
-    ONN("failed to decrypt client.p12",
-        ne_ssl_clicert_decrypt(def_cli_cert, "foobar"));
+    if (!ne_ssl_clicert_encrypted(def_cli_cert)) {
+        ne_ssl_clicert_free(def_cli_cert);
+        def_cli_cert = NULL;
+        t_context("client.p12 is not encrypted!?");
+        return FAIL;
+    }
+
+    if (ne_ssl_clicert_decrypt(def_cli_cert, P12_PASSPHRASE)) {
+        ne_ssl_clicert_free(def_cli_cert);
+        def_cli_cert = NULL;
+        t_context("failed to decrypt client.p12");
+        return FAIL;
+    }
+
+    nul_cn_fn = ne_concat(srcdir, "/nulcn.pem", NULL);
 
     return OK;
 }
@@ -294,9 +321,12 @@ static int load_client_cert(void)
     ONN("could not load client.p12", cc == NULL);
     ONN("client.p12 not encrypted!?", !ne_ssl_clicert_encrypted(cc));
     name = ne_ssl_clicert_name(cc);
-    ONN("no friendly name given", name == NULL);
-    ONV(strcmp(name, CC_NAME), ("friendly name was %s not %s", name, CC_NAME));
-    ONN("failed to decrypt", ne_ssl_clicert_decrypt(cc, "foobar"));
+    if (name == NULL) {
+        t_warning("no friendly name given");
+    } else {
+        ONV(strcmp(name, CC_NAME), ("friendly name was %s not %s", name, CC_NAME));
+    }
+    ONN("failed to decrypt", ne_ssl_clicert_decrypt(cc, P12_PASSPHRASE));
     ne_ssl_clicert_free(cc);
 
     cc = ne_ssl_clicert_read("client.p12");
@@ -325,13 +355,22 @@ static int load_client_cert(void)
     ONV(name != NULL, ("noclient.p12 had friendly name `%s'", name));
     ne_ssl_clicert_free(cc);
 
-    /* test for ccert without a friendly name, noclient.p12 */
+    /* test for ccert with a bundled CA. */
     cc = ne_ssl_clicert_read("clientca.p12");
     ONN("could not load clientca.p12", cc == NULL);
     ONN("encrypted cert marked unencrypted?", !ne_ssl_clicert_encrypted(cc));
-    ONN("could not decrypt clientca.p12", ne_ssl_clicert_decrypt(cc, "foobar"));
+    ONN("could not decrypt clientca.p12", 
+        ne_ssl_clicert_decrypt(cc, P12_PASSPHRASE));
     ne_ssl_clicert_free(cc);
 
+    /* test for ccert without a private key, nkclient.p12 */
+    cc = ne_ssl_clicert_read("nkclient.p12");
+    ONN("did not fail to load clicert without pkey", cc != NULL);
+    
+    /* test for ccert without a cert, ncclient.p12 */
+    cc = ne_ssl_clicert_read("ncclient.p12");
+    ONN("did not fail to load clicert without cert", cc != NULL);
+
     /* tests for loading bogus files. */
     cc = ne_ssl_clicert_read("Makefile");
     ONN("loaded Makefile as client cert!?", cc != NULL);
@@ -343,6 +382,23 @@ static int load_client_cert(void)
     return OK;
 }
 
+static int clicert_import(void)
+{
+    ne_ssl_client_cert *cc;
+    ne_buffer *buf = ne_buffer_create();
+
+    CALL(file_to_buffer("client.p12", buf));
+
+    cc = ne_ssl_clicert_import((unsigned char *)buf->data, ne_buffer_size(buf));
+    ONN("could not import client cert from buffer", cc == NULL);
+    
+    ONN("failed to decrypt", ne_ssl_clicert_decrypt(cc, P12_PASSPHRASE));
+    ne_ssl_clicert_free(cc);
+    ne_buffer_destroy(buf);
+
+    return OK;
+}
+
 /* Test that 'cert', which is signed by CA_CERT, is accepted
  * unconditionaly. */
 static int accept_signed_cert_for_hostname(char *cert, const char *hostname)
@@ -366,17 +422,27 @@ static int simple(void)
     return accept_signed_cert(SERVER_CERT);
 }
 
+#if 0 /* No longer works for modern SSL libraries, rightly so. */
 /* Test for SSL operation when server uses SSLv2 */
 static int simple_sslv2(void)
 {
     ne_session *sess = ne_session_create("https", "localhost", 7777);
     struct ssl_server_args args = {SERVER_CERT, 0};
+
     args.use_ssl2 = 1;
     ne_set_session_flag(sess, NE_SESSFLAG_SSLv2, 1);
+
+    if (ne_get_session_flag(sess, NE_SESSFLAG_SSLv2) != 1) {
+        t_context("no SSLv2 support in SSL library");
+        ne_session_destroy(sess);
+        return SKIP;
+    }
+
     CALL(any_ssl_request(sess, ssl_server, &args, CA_CERT, NULL, NULL));
     ne_session_destroy(sess);
     return OK;
 }
+#endif
 
 /* Serves using HTTP/1.0 get-till-EOF semantics. */
 static int serve_eof(ne_socket *sock, void *ud)
@@ -402,6 +468,15 @@ static int simple_eof(void)
     return OK;
 }
 
+static int intermediary(void)
+{
+    ne_session *sess = DEFSESS;
+    struct ssl_server_args args = {CA2_SERVER_CERT, 0};
+    CALL(any_ssl_request(sess, ssl_server, &args, CA_CERT, NULL, NULL));
+    ne_session_destroy(sess);
+    return OK;
+}
+
 static int empty_truncated_eof(void)
 {
     ne_session *sess = DEFSESS;
@@ -466,6 +541,20 @@ static int wildcard_match(void)
     return OK;
 }
 
+static int wildcard_match_altname(void)
+{
+    ne_session *sess;
+    struct ssl_server_args args = {"altname9.cert", 0};
+    
+    sess = ne_session_create("https", "anything.example.com", 443);
+    ne_session_proxy(sess, "localhost", 7777);
+
+    CALL(any_ssl_request(sess, tunnel_server, &args, CA_CERT, NULL, NULL));
+    ne_session_destroy(sess);
+    
+    return OK;
+}
+
 /* Check that hostname comparisons are not cases-sensitive. */
 static int caseless_match(void)
 {
@@ -584,12 +673,15 @@ static int parse_cert(void)
     return OK;
 }
 
+#define WRONGCN_DNAME "Bad Hostname Department, Neon Hackers Ltd, " \
+    "Cambridge, Cambridgeshire, GB"
+
 /* Check the certificate chain presented against known dnames. */
 static int check_chain(void *userdata, int fs, const ne_ssl_certificate *cert)
 {
     int *ret = userdata;
 
-    if (check_cert_dnames(cert, SERVER_DNAME, CACERT_DNAME) == FAIL) {
+    if (check_cert_dnames(cert, WRONGCN_DNAME, CACERT_DNAME) == FAIL) {
         *ret = -1;
         return 0;
     }
@@ -615,13 +707,13 @@ static int parse_chain(void)
 {
     ne_session *sess = DEFSESS;
     int ret = 0;
-    struct ssl_server_args args = {SERVER_CERT, 0};
+    struct ssl_server_args args = {"wrongcn.cert", 0};
 
-    args.ca_list = "ca/cert.pem";    
+    args.ca_list = CA_CERT;
 
-    /* don't give a CA cert; should force the verify callback to be
-     * used. */
-    CALL(any_ssl_request(sess, ssl_server, &args, NULL, 
+    /* The cert is signed by the CA but has a CN mismatch, so will
+     * force the verification callback to be invoked. */
+    CALL(any_ssl_request(sess, ssl_server, &args, CA_CERT, 
 			 check_chain, &ret));
     ne_session_destroy(sess);
 
@@ -660,15 +752,15 @@ static int no_verify(void)
 static int cache_verify(void)
 {
     ne_session *sess = DEFSESS;
-    int ret, count = 0;
+    int count = 0;
     struct ssl_server_args args = {SERVER_CERT, 0};
     
     /* force verify cert. */
-    ret = any_ssl_request(sess, ssl_server, &args, NULL, count_vfy,
-			  &count);
+    CALL(any_ssl_request(sess, ssl_server, &args, NULL, count_vfy,
+                         &count));
 
     CALL(spawn_server(7777, ssl_server, &args));
-    ret = any_request(sess, "/foo2");
+    ONREQ(any_request(sess, "/foo2"));
     CALL(await_server());
 
     ONV(count != 1,
@@ -687,20 +779,62 @@ static int get_failures(void *userdata, int fs, const ne_ssl_certificate *c)
     return -1;
 }
 
-/* Helper function: run a request using the given self-signed server
- * certificate, and expect the request to fail with the given
- * verification failure flags. */
-static int fail_ssl_request(char *cert, char *cacert, const char *host,
-			    const char *msg, int failures)
+/* Helper function for expected-to-fail SSL tests.
+ *
+ * An SSL server is spawned using 'cert' and 'key' as the key pair.
+ * The client will trust CA cert 'cacert', and use 'host' as the server
+ * name.  If realhost is non-NULL, this address will be used to connect
+ * to in favour of host; the server is otherwise identified as 'host'.
+ * 'msg' must be a substring of the error string.
+ * 'failures' must equal the failure bitmask passed to the verify
+ * callback in the client.
+ * If none of the expected conditions is met, 'errstr' will be
+ * used in the test failure context.
+ */
+static int fail_ssl_request_with_error2(char *cert, char *key, char *cacert, 
+                                        const char *host, const char *realhost,
+                                        const char *msg, int failures,
+                                        const char *errstr)
 {
     ne_session *sess = ne_session_create("https", host, 7777);
     int gotf = 0, ret;
+    struct ssl_server_args args = {0};
+    ne_sock_addr *addr = NULL;
+    const ne_inet_addr **list = NULL;
 
-    ret = any_ssl_request(sess, fail_serve, cert, cacert,
+    if (realhost) {
+        size_t n;
+        const ne_inet_addr *ia;
+
+        addr = ne_addr_resolve(realhost, 0);
+
+        ONV(ne_addr_result(addr),
+            ("fake hostname lookup failed for %s", realhost));
+
+        NE_DEBUG(NE_DBG_SSL, "ssl: Using fake hostname '%s'\n", realhost);
+
+        for (n = 0, ia = ne_addr_first(addr); ia; ia = ne_addr_next(addr))
+            n++;
+
+        NE_DEBUG(NE_DBG_SSL, "ssl: Address count '%lu'\n", n);
+
+        list = ne_calloc(n * sizeof(*list));
+
+        for (n = 0, ia = ne_addr_first(addr); ia; ia = ne_addr_next(addr))
+            list[n++] = ia;
+        
+        ne_set_addrlist(sess, list, n);
+    }
+
+    args.cert = cert;
+    args.key = key;
+    args.fail_silently = 1;
+    
+    ret = any_ssl_request(sess, ssl_server, &args, cacert,
 			  get_failures, &gotf);
 
     ONV(gotf == 0,
-	("no error in verification callback; request failed: %s",
+	("no error in verification callback; error string: %s",
 	 ne_get_error(sess)));
 
     ONV(gotf & ~NE_SSL_FAILMASK,
@@ -713,11 +847,38 @@ static int fail_ssl_request(char *cert, char *cacert, const char *host,
     /* and check that the request was failed too. */
     ONV(ret == NE_OK, ("%s", msg));
 
+    ONV(errstr && strstr(ne_get_error(sess), errstr) == NULL,
+        ("unexpected failure message '%s', wanted '%s'",
+         ne_get_error(sess), errstr));
+        
     ne_session_destroy(sess);
+    if (addr) ne_addr_destroy(addr);
 
     return OK;
 }
 
+/* Helper function: run a request using the given self-signed server
+ * certificate, and expect the request to fail with the given
+ * verification failure flags. */
+static int fail_ssl_request_with_error(char *cert, char *cacert, const char *host,
+                                       const char *msg, int failures,
+                                       const char *errstr)
+{
+    return fail_ssl_request_with_error2(cert, NULL, cacert, host, NULL,
+                                        msg, failures, errstr);
+}
+
+
+/* Helper function: run a request using the given self-signed server
+ * certificate, and expect the request to fail with the given
+ * verification failure flags. */
+static int fail_ssl_request(char *cert, char *cacert, const char *host,
+			    const char *msg, int failures)
+{
+    return fail_ssl_request_with_error(cert, cacert, host, msg, failures,
+                                       NULL);
+}        
+
 /* Note that the certs used for fail_* are mostly self-signed, so the
  * cert is passed as CA cert and server cert to fail_ssl_request. */
 
@@ -725,17 +886,51 @@ static int fail_ssl_request(char *cert, char *cacert, const char *host,
  * flagged as such. */
 static int fail_wrongCN(void)
 {
-    return fail_ssl_request("wrongcn.pem", "wrongcn.pem", "localhost",
-			    "certificate with incorrect CN was accepted",
-			    NE_SSL_IDMISMATCH);
+    return fail_ssl_request_with_error("wrongcn.cert", "ca/cert.pem", "localhost",
+                                       "certificate with incorrect CN was accepted",
+                                       NE_SSL_IDMISMATCH,
+                                       "certificate issued for a different hostname");
+                            
+}
+
+#define SRCDIR(s) ne_concat(srcdir, "/" s, NULL)
+
+static int fail_nul_cn(void)
+{
+    char *key = SRCDIR("nulsrv.key"), *ca = SRCDIR("nulca.pem");
+    CALL(fail_ssl_request_with_error2(nul_cn_fn, key, ca,
+                                      "www.bank.com", "localhost",
+                                      "certificate with incorrect CN was accepted",
+                                      NE_SSL_IDMISMATCH,
+                                      "certificate issued for a different hostname"));
+    ne_free(key);
+    ne_free(ca);
+    return OK;
+}
+
+static int fail_nul_san(void)
+{
+    char *cert = SRCDIR("nulsan.pem"), *key = SRCDIR("nulsrv.key"),
+        *ca = SRCDIR("nulca.pem");
+    CALL(fail_ssl_request_with_error2(cert, key, ca, 
+                                      "www.bank.com", "localhost",
+                                      "certificate with incorrect CN was accepted",
+                                      NE_SSL_IDMISMATCH,
+                                      "certificate issued for a different hostname"));
+    ne_free(cert);
+    ne_free(key);
+    ne_free(ca);
+    return OK;
 }
 
 /* Check that an expired certificate is flagged as such. */
 static int fail_expired(void)
 {
     char *c = ne_concat(srcdir, "/expired.pem", NULL);
-    CALL(fail_ssl_request(c, c,  "localhost",
-                          "expired certificate was accepted", NE_SSL_EXPIRED));
+    CALL(fail_ssl_request_with_error(c, c,  "localhost",
+                                     "expired certificate was accepted", 
+                                     NE_SSL_EXPIRED,
+                                     "certificate has expired"));
     ne_free(c);
     return OK;
 }
@@ -743,9 +938,10 @@ static int fail_expired(void)
 static int fail_notvalid(void)
 {
     char *c = ne_concat(srcdir, "/notvalid.pem", NULL);
-    CALL(fail_ssl_request(c, c,  "localhost",
-                          "not yet valid certificate was accepted",
-                          NE_SSL_NOTYETVALID));
+    CALL(fail_ssl_request_with_error(c, c,  "localhost",
+                                     "not yet valid certificate was accepted",
+                                     NE_SSL_NOTYETVALID,
+                                     "certificate is not yet valid"));
     ne_free(c);
     return OK;    
 }
@@ -754,8 +950,9 @@ static int fail_notvalid(void)
  * fail with UNTRUSTED. */
 static int fail_untrusted_ca(void)
 {
-    return fail_ssl_request("server.cert", NULL, "localhost",
-                            "untrusted CA.", NE_SSL_UNTRUSTED);
+    return fail_ssl_request_with_error("server.cert", NULL, "localhost",
+                                       "untrusted CA.", NE_SSL_UNTRUSTED,
+                                       "issuer is not trusted");
 }
 
 static int fail_self_signed(void)
@@ -802,6 +999,32 @@ static int fail_bad_urialtname(void)
                             "bad URI altname cert", NE_SSL_IDMISMATCH);
 }
 
+static int fail_wildcard(void)
+{
+    return fail_ssl_request("altname9.cert", CA_CERT, "localhost",
+                            "subjaltname not honored", NE_SSL_IDMISMATCH);
+}
+
+static int fail_wildcard_ip(void)
+{
+    return fail_ssl_request("wildip.cert", CA_CERT, "127.0.0.1",
+                            "wildcard IP", NE_SSL_IDMISMATCH);
+}
+
+static int fail_ca_expired(void)
+{
+    return fail_ssl_request_with_error("ca1server.cert", "ca1/cert.pem", 
+                                       "localhost", "issuer ca expired", 
+                                       NE_SSL_BADCHAIN,
+                                       "bad certificate chain");
+}
+
+static int fail_ca_notyetvalid(void)
+{
+    return fail_ssl_request("ca3server.cert", "ca3/cert.pem", "localhost",
+                            "issuer ca not yet valid", NE_SSL_BADCHAIN);
+}
+
 /* Test that the SSL session is cached across connections. */
 static int session_cache(void)
 {
@@ -849,7 +1072,7 @@ static int client_cert_provided(void)
     cc = ne_ssl_clicert_read("client.p12");
     ONN("could not load client.p12", cc == NULL);
     ONN("could not decrypt client.p12", 
-        ne_ssl_clicert_decrypt(cc, "foobar"));
+        ne_ssl_clicert_decrypt(cc, P12_PASSPHRASE));
     
     ne_ssl_provide_clicert(sess, ccert_provider, cc);
     CALL(any_ssl_request(sess, ssl_server, &args, CA_CERT,
@@ -860,11 +1083,14 @@ static int client_cert_provided(void)
     return OK;
 }
 
+#define DN_COUNT 5
+
 static void cc_check_dnames(void *userdata, ne_session *sess,
                             const ne_ssl_dname *const *dns, int dncount)
 {
     int n, *ret = userdata;
-    static const char *expected[4] = {
+    static const char *expected[DN_COUNT] = {
+        CACERT_DNAME,
         "First Random CA, CAs Ltd., Lincoln, Lincolnshire, GB",
         "Second Random CA, CAs Ltd., Falmouth, Cornwall, GB",
         "Third Random CA, CAs Ltd., Ipswich, Suffolk, GB",
@@ -873,13 +1099,14 @@ static void cc_check_dnames(void *userdata, ne_session *sess,
 
     ne_ssl_set_clicert(sess, def_cli_cert);
 
-    if (dncount != 4) {
-        t_context("dname count was %d not 4", dncount);
+    if (dncount != DN_COUNT) {
+        t_context("dname count was %d not %d", dncount, 
+                  DN_COUNT);
         *ret = -1;
         return;
     }
     
-    for (n = 0; n < 4; n++) {
+    for (n = 0; n < DN_COUNT; n++) {
         char which[5];
 
         sprintf(which, "%d", n);
@@ -901,7 +1128,7 @@ static int cc_provided_dnames(void)
     struct ssl_server_args args = {SERVER_CERT, NULL};
 
     args.require_cc = 1;
-    args.send_ca = "calist.pem";
+    args.ca_list = "calist.pem";
 
     PRECOND(def_cli_cert);
 
@@ -933,6 +1160,30 @@ static int client_cert_pkcs12(void)
     return OK;
 }
 
+/* Test use of a PKCS#12 cert with an embedded CA cert - fails with <=
+ * 0.28.3 in GnuTLS build. */
+static int client_cert_ca(void)
+{
+    ne_session *sess = DEFSESS;
+    struct ssl_server_args args = {SERVER_CERT, NULL};
+    ne_ssl_client_cert *cc;
+
+    args.require_cc = 1;
+
+    cc = ne_ssl_clicert_read("clientca.p12");
+    ONN("could not load clientca.p12", cc == NULL);
+    ONN("encrypted cert marked unencrypted?", !ne_ssl_clicert_encrypted(cc));
+    ONN("could not decrypt clientca.p12", 
+        ne_ssl_clicert_decrypt(cc, P12_PASSPHRASE));
+
+    ne_ssl_set_clicert(sess, cc);
+    CALL(any_ssl_request(sess, ssl_server, &args, CA_CERT, NULL, NULL));
+
+    ne_ssl_clicert_free(cc);
+
+    ne_session_destroy(sess);    
+    return OK;
+}
 
 /* Tests use of an unencrypted client certificate. */
 static int ccert_unencrypted(void)
@@ -955,6 +1206,38 @@ static int ccert_unencrypted(void)
     return OK;
 }
 
+#define NOCERT_MESSAGE "client certificate was requested"
+
+/* Tests for useful error message if a handshake fails where a client
+ * cert was requested. */
+static int no_client_cert(void)
+{
+    ne_session *sess = DEFSESS;
+    struct ssl_server_args args = {SERVER_CERT, NULL};
+    int ret;
+
+    args.require_cc = 1;
+    args.fail_silently = 1;
+
+    ne_ssl_trust_cert(sess, def_ca_cert);
+
+    CALL(spawn_server(7777, ssl_server, &args));
+    
+    ret = any_request(sess, "/failme");
+
+    ONV(ret != NE_ERROR,
+        ("unexpected result %d: %s", ret, ne_get_error(sess)));
+
+    ONV(strstr(ne_get_error(sess), NOCERT_MESSAGE) == NULL,
+        ("error message was '%s', missing '%s'", 
+         ne_get_error(sess), NOCERT_MESSAGE));
+    
+    reap_server();
+
+    ne_session_destroy(sess);    
+    return OK;
+}
+
 /* non-zero if a server auth header was received */
 static int got_server_auth; 
 
@@ -1202,6 +1485,24 @@ static int cert_identities(void)
     return OK;
 }
 
+static int nulcn_identity(void)
+{
+    ne_ssl_certificate *cert = ne_ssl_cert_read(nul_cn_fn);
+    const char *id, *expected = "www.bank.com\\x00.badguy.com";
+
+    ONN("could not read nulcn.pem", cert == NULL);
+
+    id = ne_ssl_cert_identity(cert);
+
+    ONV(id != NULL
+        && strcmp(id, expected) != 0,
+        ("certificate `nulcn.pem' had identity `%s' not `%s'", 
+         id, expected));
+    
+    ne_ssl_cert_free(cert);
+    return OK;
+}
+
 static int check_validity(const char *fname,
                           const char *from, const char *until)
 {
@@ -1502,6 +1803,70 @@ static int nonssl_trust(void)
     return OK;
 }
 
+/* PIN password provider callback. */
+static int pkcs11_pin(void *userdata, int attempt,
+                      const char *slot_descr, const char *token_label,
+                      unsigned int flags, char *pin)
+{
+    char *sekrit = userdata;
+
+    NE_DEBUG(NE_DBG_SSL, "pkcs11: slot = [%s], token = [%s]\n",
+             slot_descr, token_label);
+
+    if (attempt == 0) {
+        strcpy(pin, sekrit);
+        return 0;
+    }
+    else {
+        return -1;
+    }
+}
+
+static int nss_pkcs11_test(const char *dbname)
+{
+    ne_session *sess = DEFSESS;
+    struct ssl_server_args args = {SERVER_CERT, NULL};
+    ne_ssl_pkcs11_provider *prov;
+    int ret;
+
+    args.require_cc = 1;
+
+    if (access(dbname, R_OK|X_OK)) {
+        t_warning("NSS required for PKCS#11 testing");
+        return SKIP;
+    }
+
+    ret = ne_ssl_pkcs11_nss_provider_init(&prov, "softokn3", dbname, NULL, 
+                                          NULL, NULL);
+    if (ret) {
+        if (ret == NE_PK11_NOTIMPL)
+            t_context("pakchois library required for PKCS#11 support");
+        else
+            t_context("could not load NSS softokn3 PKCS#11 provider");
+        return SKIP;
+    }
+
+    ne_ssl_pkcs11_provider_pin(prov, pkcs11_pin, "foobar");
+    ne_ssl_set_pkcs11_provider(sess, prov);
+
+    ret = any_ssl_request(sess, ssl_server, &args, CA_CERT, NULL, NULL);
+
+    ne_session_destroy(sess);
+    ne_ssl_pkcs11_provider_destroy(prov);
+
+    return ret;
+}
+
+static int pkcs11(void)
+{
+    return nss_pkcs11_test("nssdb");
+}
+
+static int pkcs11_dsa(void)
+{
+    return nss_pkcs11_test("nssdb-dsa");
+}
+
 /* TODO: code paths still to test in cert verification:
  * - server cert changes between connections: Mozilla gives
  * a "bad MAC decode" error for this; can do better?
@@ -1532,18 +1897,24 @@ ne_test tests[] = {
     T(read_write),
 
     T(load_client_cert),
+    T(clicert_import),
 
     T(simple),
+#if 0
     T(simple_sslv2),
+#endif
     T(simple_eof),
     T(empty_truncated_eof),
     T(fail_not_ssl),
     T(cache_cert),
+    T(intermediary),
 
     T(client_cert_pkcs12),
     T(ccert_unencrypted),
     T(client_cert_provided),
     T(cc_provided_dnames),
+    T(no_client_cert),
+    T(client_cert_ca),
 
     T(parse_cert),
     T(parse_chain),
@@ -1551,6 +1922,7 @@ ne_test tests[] = {
     T(no_verify),
     T(cache_verify),
     T(wildcard_match),
+    T(wildcard_match_altname),
     T(caseless_match),
 
     T(subject_altname),
@@ -1572,7 +1944,15 @@ ne_test tests[] = {
     T(fail_host_ipaltname),
     T(fail_bad_ipaltname),
     T(fail_bad_urialtname),
+    T(fail_wildcard),
+    T(fail_wildcard_ip),
+    T(fail_ca_notyetvalid),
+    T(fail_ca_expired),
 
+    T(nulcn_identity),
+    T(fail_nul_cn),
+    T(fail_nul_san),
+    
     T(session_cache),
 	
     T(fail_tunnel),
@@ -1583,5 +1963,8 @@ ne_test tests[] = {
 
     T(nonssl_trust),
 
+    T(pkcs11),
+    T_XFAIL(pkcs11_dsa), /* unclear why this fails currently. */
+
     T(NULL) 
 };
