@@ -222,7 +222,7 @@ static int cmp(void)
             ("'%s' and '%s' did not compare as different",
              diff[n].left, diff[n].right));
 
-        ONV(r1 + r2 != 0,
+        ONV(((r1 > 0) != (r2 < 0) || (r1 < 0) != (r2 > 0)),
             ("'%s' and '%s' did not compare reflexively (%d vs %d)",
              diff[n].left, diff[n].right, r1, r2));
 
@@ -330,6 +330,9 @@ static int parse(void)
         { "foo/bar", NULL, NULL, 0, "foo/bar", NULL, NULL, NULL }, /* path-noscheme */
         { "", NULL, NULL, 0, "", NULL, NULL, NULL }, /* path-empty */
 
+        /* CVE-2007-0157: buffer under-read in 0.26.[012]. */
+        { "http://webdav.org\xFF", "http", "webdav.org\xFF", 0, "/", NULL, NULL, NULL },
+
 	{ NULL }
     };
     int n;
@@ -506,6 +509,8 @@ static int resolve(void)
         ONN("bad pointer was returned", 
             ne_uri_resolve(&base, &relative, &resolved) != &resolved);
 
+        ONN("NULL path after resolve", resolved.path == NULL);
+
         actual = ne_uri_unparse(&resolved);
         
         ONCMP(ts[n].expected, actual, ts[n].relative, "output mismatch");
